from collections import OrderedDict
from django.core.cache import cache

from .signals import setting_changed

__all__ = ['get_all_settings', 'get_setting', 'get_setting_storage',
           'register_setting', 'unregister_setting', 'set_setting_value']


_settings = OrderedDict()


def _get_cache_key(module_name, class_name, attribute_name):
    return '.'.join(['dbsettings', module_name, class_name, attribute_name])


def get_all_settings():
    return list(_settings.values())


def get_app_settings(app_label):
    return [p for p in _settings.values() if app_label == p.app]


def get_setting(module_name, class_name, attribute_name):
    return _settings[module_name, class_name, attribute_name]


def setting_in_db(module_name, class_name, attribute_name):
    from dbsettings.models import Setting
    return Setting.objects.filter(
        module_name=module_name,
        class_name=class_name,
        attribute_name=attribute_name,
    ).count() == 1


def get_setting_storage(module_name, class_name, attribute_name):
    from dbsettings.models import Setting
    from dbsettings.settings import USE_CACHE, CACHE_EXPIRATION
    storage = None
    if USE_CACHE:
        key = _get_cache_key(module_name, class_name, attribute_name)
        try:
            storage = cache.get(key)
        except:
            pass
    if storage is None:
        try:
            storage = Setting.objects.get(
                module_name=module_name,
                class_name=class_name,
                attribute_name=attribute_name,
            )
        except Setting.DoesNotExist:
            setting_object = get_setting(module_name, class_name, attribute_name)
            storage = Setting(
                module_name=module_name,
                class_name=class_name,
                attribute_name=attribute_name,
                value=setting_object.default,
            )
        if USE_CACHE:
            try:
                args = []
                if CACHE_EXPIRATION != -1:
                    args.append(CACHE_EXPIRATION)
                cache.set(key, storage, *args)
            except:
                pass
    return storage


def register_setting(setting):
    if setting.key not in _settings:
        _settings[setting.key] = setting


def unregister_setting(setting):
    if setting.key in _settings and _settings[setting.key] is setting:
        del _settings[setting.key]


def set_setting_value(module_name, class_name, attribute_name, value):
    from dbsettings.settings import USE_CACHE
    setting = get_setting(module_name, class_name, attribute_name)
    storage = get_setting_storage(module_name, class_name, attribute_name)
    storage.value = setting.get_db_prep_save(value, oldvalue=storage.value)
    storage.save()
    setting_changed.send(sender=setting, value=setting.to_python(value))
    if USE_CACHE:
        key = _get_cache_key(module_name, class_name, attribute_name)
        cache.delete(key)



from django.apps import apps
from django.conf import settings


sites_installed = apps.is_installed('django.contrib.sites')
USE_SITES = getattr(settings, 'DBSETTINGS_USE_SITES', sites_installed)
USE_CACHE = getattr(settings, 'DBSETTINGS_USE_CACHE', True)
CACHE_EXPIRATION = getattr(settings, 'DBSETTINGS_CACHE_EXPIRATION', -1)
VALUE_LENGTH = getattr(settings, 'DBSETTINGS_VALUE_LENGTH', 255)



from django.urls import path

from dbsettings.views import site_settings, app_settings


urlpatterns = [
    path('', site_settings, name='site_settings'),
    path('<app_label>/', app_settings, name='app_settings'),
]



import re

from collections import OrderedDict
from django.apps import apps
from django import forms
from django.utils.text import capfirst

from dbsettings.loading import get_setting_storage


RE_FIELD_NAME = re.compile(r'^(.+)__(.*)__(.+)$')


class SettingsEditor(forms.BaseForm):
    "Base editor, from which customized forms are created"

    def __iter__(self):
        for field in super(SettingsEditor, self).__iter__():
            yield self.specialize(field)

    def __getitem__(self, name):
        field = super(SettingsEditor, self).__getitem__(name)
        return self.specialize(field)

    def specialize(self, field):
        "Wrapper to add module_name and class_name for regrouping"
        field.label = capfirst(field.label)
        module_name, class_name, _ = RE_FIELD_NAME.match(field.name).groups()

        app_label = self.apps[field.name]
        field.module_name = app_label

        if class_name:
            model = apps.get_model(app_label, class_name)
            if model:
                class_name = model._meta.verbose_name
        field.class_name = class_name
        field.verbose_name = self.verbose_names[field.name]

        return field


def customized_editor(user, settings):
    "Customize the setting editor based on the current user and setting list"
    base_fields = OrderedDict()
    verbose_names = {}
    apps = {}
    for setting in settings:
        perm = '%s.can_edit_%s_settings' % (
            setting.app,
            setting.class_name.lower()
        )
        if user.has_perm(perm):
            # Add the field to the customized field list
            storage = get_setting_storage(*setting.key)
            kwargs = {
                'label': setting.description,
                'help_text': setting.help_text,
                # Provide current setting values for initializing the form
                'initial': setting.to_editor(storage.value),
                'required': setting.required,
                'widget': setting.widget,
            }
            if setting.choices:
                field = forms.ChoiceField(choices=setting.choices, **kwargs)
            else:
                field = setting.field(**kwargs)
            key = '%s__%s__%s' % setting.key
            apps[key] = setting.app
            base_fields[key] = field
            verbose_names[key] = setting.verbose_name
    attrs = {'base_fields': base_fields, 'verbose_names': verbose_names, 'apps': apps}
    return type('SettingsEditor', (SettingsEditor,), attrs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations

from dbsettings.settings import USE_SITES, VALUE_LENGTH


class Migration(migrations.Migration):

    dependencies = []
    if USE_SITES:
        dependencies.append(('sites', '0001_initial'))

    operations = [
        migrations.CreateModel(
            name='Setting',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('module_name', models.CharField(max_length=255)),
                ('class_name', models.CharField(max_length=255, blank=True)),
                ('attribute_name', models.CharField(max_length=255)),
                ('value', models.CharField(max_length=VALUE_LENGTH, blank=True)),
            ] + ([('site', models.ForeignKey(to='sites.Site', on_delete=models.CASCADE))] if USE_SITES else [])
        ),
    ]



from django.db import models

from dbsettings.settings import USE_SITES, VALUE_LENGTH

if USE_SITES:
    from django.contrib.sites.models import Site

    class SiteSettingManager(models.Manager):
        def get_queryset(self):
            sup = super(SiteSettingManager, self)
            qs = sup.get_queryset() if hasattr(sup, 'get_queryset') else sup.get_query_set()
            return qs.filter(site=Site.objects.get_current())
        get_query_set = get_queryset


class Setting(models.Model):
    module_name = models.CharField(max_length=255)
    class_name = models.CharField(max_length=255, blank=True)
    attribute_name = models.CharField(max_length=255)
    value = models.CharField(max_length=VALUE_LENGTH, blank=True)

    if USE_SITES:
        site = models.ForeignKey(Site, on_delete=models.CASCADE)
        objects = SiteSettingManager()

        def save(self, *args, **kwargs):
            self.site = Site.objects.get_current()
            return super(Setting, self).save(*args, **kwargs)

    def __bool__(self):
        return self.pk is not None



from __future__ import unicode_literals

from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.contrib.admin.views.decorators import staff_member_required
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _
from django.contrib import messages

from dbsettings import loading, forms


@staff_member_required
def app_settings(request, app_label, template='dbsettings/app_settings.html'):
    # Determine what set of settings this editor is used for
    if app_label is None:
        settings = loading.get_all_settings()
        title = _('Site settings')
    else:
        settings = loading.get_app_settings(app_label)
        title = _('%(app)s settings') % {'app': capfirst(app_label)}

    # Create an editor customized for the current user
    editor = forms.customized_editor(request.user, settings)

    if request.method == 'POST':
        # Populate the form with user-submitted data
        form = editor(request.POST.copy(), request.FILES)
        if form.is_valid():
            form.full_clean()

            for name, value in form.cleaned_data.items():
                key = forms.RE_FIELD_NAME.match(name).groups()
                setting = loading.get_setting(*key)
                try:
                    storage = loading.get_setting_storage(*key)
                    current_value = setting.to_python(storage.value)
                except:
                    current_value = None

                if current_value != setting.to_python(value):
                    args = key + (value,)
                    loading.set_setting_value(*args)

                    # Give user feedback as to which settings were changed
                    if setting.class_name:
                        location = setting.class_name
                    else:
                        location = setting.module_name
                    update_msg = (_('Updated %(desc)s on %(location)s') %
                                  {'desc': str(setting.description),
                                   'location': location})
                    messages.add_message(request, messages.INFO, update_msg)

            return HttpResponseRedirect(request.path)
    else:
        # Leave the form populated with current setting values
        form = editor()

    return render(request, template, {
        'title': title,
        'form': form,
    })


# Site-wide setting editor is identical, but without an app_label
def site_settings(request):
    return app_settings(request, app_label=None, template='dbsettings/site_settings.html')
# staff_member_required is implied, since it calls app_settings



from django import VERSION
from django.db.models.signals import post_migrate


def mk_permissions(permissions, appname, verbosity):
    """
    Make permission at app level - hack with empty ContentType.

    Adapted code from http://djangosnippets.org/snippets/334/
    """
    from django.contrib.auth.models import Permission
    from django.contrib.contenttypes.models import ContentType
    # create a content type for the app
    defaults = {} if VERSION >= (1, 10) else {'name': appname}
    ct, created = ContentType.objects.get_or_create(model='', app_label=appname,
                                                    defaults=defaults)
    if created and verbosity >= 2:
        print("Adding custom content type '%s'" % ct)
    # create permissions
    for codename, name in permissions:
        p, created = Permission.objects.get_or_create(codename=codename,
                                                      content_type__pk=ct.id,
                                                      defaults={'name': name, 'content_type': ct})
        if created and verbosity >= 2:
            print("Adding custom permission '%s'" % p)


def handler(sender, **kwargs):
    from dbsettings.loading import get_app_settings
    app_label = sender.label
    are_global_settings = any(not s.class_name for s in get_app_settings(app_label))
    if are_global_settings:
        permission = (
            'can_edit__settings',
            'Can edit %s non-model settings' % app_label,
        )
        mk_permissions([permission], app_label, 0)


post_migrate.connect(handler)
